
BEGIN {
  unless ($ENV{AUTHOR_TESTING}) {
    print qq{1..0 # SKIP these tests are for testing by the author\n};
    exit
  }
}

use strict;
use warnings;
use Test::More;

# generated by Dist::Zilla::Plugin::Test::PodSpelling 2.007005
use Test::Spelling 0.12;
use Pod::Wordlist;


add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( bin lib ) );
__DATA__
AMQP
AnyEvent
Carl
Channel
Dave
Doran
Hörberg
Ikuta
José
Julio
Lambley
LocalQueue
Masahito
Micó
Neil
Nicolas
Polo
RabbitMQ
Ragusa
Rod
Ruslan
Scott
TCP
Taylor
Tom
Zakirov
ack
aragusa
atoomic
bobtfish
carl
cooldaemon
developers
dlambley
julio
lib
qos
rod
ruz
scott
